/*
 * Copyright (c) 2019 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package main

//go:generate gogitversion -p main

import (
	"encoding/hex"
	"flag"
	"fmt"
	"log"
	"math/rand"
	"net"
	"net/http"
	"os"
	"strings"
	"time"
)

var (
	hostname  string
	logger    *log.Logger
	address   = flag.String("address", ":7357", "listen address")
	logF      = flag.String("log", "", "log file (default stdout)")
	setPrefix = flag.String("set_prefix", "X-Set-",
		"request header prefix for setting response headers")
	versionF = flag.Bool("version", false, "show version and exit")
)

const tFormat = "02/Jan/2006:15:04:05 -0700"

func reqID() string {
	bytes := make([]byte, 16)
	rand.Read(bytes)
	return hex.EncodeToString(bytes)
}

func reqLog(req *http.Request, t, id string) {
	clientIP := req.RemoteAddr
	if colon := strings.LastIndex(clientIP, ":"); colon != -1 {
		clientIP = clientIP[:colon]
	}
	logger.Printf("%s - - [%s] \"%s %s %s\" 200 (%s)\n", clientIP,
		t, req.Method, req.RequestURI, req.Proto, id)
}

func handle(resp http.ResponseWriter, req *http.Request) {
	now := time.Now().Format(tFormat)
	id := reqID()

	respHdr := resp.Header()
	respHdr.Set("Content-Type", "text/plain")

	for reqHdr, vals := range req.Header {
		if !strings.HasPrefix(reqHdr, *setPrefix) {
			continue
		}
		newHdr := string([]byte(reqHdr)[len(*setPrefix):])
		for _, val := range vals {
			respHdr.Add(newHdr, val)
		}
	}

	addrVal := req.Context().Value(http.LocalAddrContextKey)
	if addrVal == nil {
		fmt.Fprintln(resp, "Server address: unknown")
	} else if addr, ok := addrVal.(net.Addr); !ok {
		fmt.Fprintln(resp, "Server address: unknown")
	} else {
		fmt.Fprintln(resp, "Server address:", addr.String())
	}
	fmt.Fprintln(resp, "Server name:", hostname)
	fmt.Fprintln(resp, "Date:", now)
	fmt.Fprintln(resp, "URI:", req.RequestURI)
	fmt.Fprintln(resp, "Request ID:", id)
	fmt.Fprintln(resp)
	req.Write(resp)
	reqLog(req, now, id)
}

func main() {
	var err error

	flag.Parse()

	if *versionF {
		fmt.Println("http-echo version", version)
		os.Exit(0)
	}

	out := os.Stdout
	if *logF != "" && *logF != "-" {
		if out, err = os.OpenFile(*logF,
			os.O_CREATE|os.O_WRONLY|os.O_APPEND, 0644); err != nil {
			log.Fatalf("%v", err)
		}
	}
	logger = log.New(out, "", log.LstdFlags)
	logger.Println("starting http-echo version", version)

	if hostname, err = os.Hostname(); err != nil {
		logger.Printf("cannot get hostname: %v", err)
		hostname = "unknown"
	}

	mux := http.NewServeMux()
	mux.HandleFunc("/", handle)
	logger.Println("listening at:", *address)
	logger.SetFlags(0)
	logger.Fatal(http.ListenAndServe(*address, mux))
}
